import os
from dataclasses import dataclass

# Try to load environment variables from .env file
# This is mainly for development; in production, env vars should be set
# in the environment
if os.environ.get("DEV_ENV", "1") == "1":
    try:
        from dotenv import load_dotenv

        load_dotenv()
    except ImportError:
        # python-dotenv is not installed, continue without it
        # environment variables will be read directly from the OS
        pass

from airunner.utils.settings.get_qsettings import get_qsettings
from airunner.settings import (
    AIRUNNER_LLM_PERFORM_CONVERSATION_SUMMARY,
    AIRUNNER_LLM_USE_WEATHER_PROMPT,
    AIRUNNER_LLM_AGENT_UPDATE_MOOD_AFTER_N_TURNS,
    AIRUNNER_LLM_AGENT_MAX_FUNCTION_CALLS,
    AIRUNNER_LLM_AGENT_SUMMARIZE_AFTER_N_TURNS,
    AIRUNNER_LLM_OPENROUTER_MODEL,
    AIRUNNER_LLM_PRINT_SYSTEM_PROMPT,
    AIRUNNER_LLM_PERFORM_ANALYSIS,
    AIRUNNER_LLM_UPDATE_USER_DATA_ENABLED,
    AIRUNNER_LLM_USE_CHATBOT_MOOD,
    AIRUNNER_LLM_PERFORM_CONVERSATION_RAG,
)


@dataclass
class LLMSettings:
    """
    Settings for Large Language Model functionality.
    Contains configuration options for local and API-based LLM usage.
    """

    use_weather_prompt: bool = (
        AIRUNNER_LLM_USE_WEATHER_PROMPT  # Fixed trailing comma
    )
    update_mood_after_n_turns: int = (
        AIRUNNER_LLM_AGENT_UPDATE_MOOD_AFTER_N_TURNS
    )
    summarize_after_n_turns: int = AIRUNNER_LLM_AGENT_SUMMARIZE_AFTER_N_TURNS
    perform_conversation_summary: bool = (
        AIRUNNER_LLM_PERFORM_CONVERSATION_SUMMARY
    )
    max_function_calls: int = AIRUNNER_LLM_AGENT_MAX_FUNCTION_CALLS
    model: str = AIRUNNER_LLM_OPENROUTER_MODEL
    print_llm_system_prompt: bool = AIRUNNER_LLM_PRINT_SYSTEM_PROMPT
    llm_perform_analysis: bool = AIRUNNER_LLM_PERFORM_ANALYSIS
    update_user_data_enabled: bool = AIRUNNER_LLM_UPDATE_USER_DATA_ENABLED
    use_chatbot_mood: bool = AIRUNNER_LLM_USE_CHATBOT_MOOD
    perform_conversation_rag: bool = AIRUNNER_LLM_PERFORM_CONVERSATION_RAG



import numpy as np
from PIL import Image

from airunner.filters.base_filter import BaseFilter


class Dither(BaseFilter):
    """Apply dithering to an image.
    
    This filter applies Floyd-Steinberg dithering to create a black 
    and white image with the illusion of more tones.
    
    Attributes:
        threshold: Value between 0 and 1 determining the threshold for
                 converting pixels to black or white.
    """
    
    def apply_filter(self, image, do_reset=False):
        """Apply dithering filter to the image.
        
        Args:
            image: The PIL Image to filter.
            do_reset: Whether to reset internal state (unused in this filter).
            
        Returns:
            The dithered black and white PIL Image.
        """
        # Convert image to grayscale
        grayscale_image = image.convert('L')
        
        # Convert grayscale image to numpy array
        image_array = np.array(grayscale_image, dtype=float)
        
        # Calculate the threshold value
        max_value = 255
        threshold_value = self.threshold * max_value
        
        # Apply Floyd-Steinberg dithering
        for y in range(image_array.shape[0]):
            for x in range(image_array.shape[1]):
                old_pixel = image_array[y, x]
                new_pixel = 255 if old_pixel > threshold_value else 0
                image_array[y, x] = new_pixel
                quant_error = old_pixel - new_pixel
                
                if x + 1 < image_array.shape[1]:
                    image_array[y, x + 1] += quant_error * 7 / 16
                
                if y + 1 < image_array.shape[0]:
                    if x > 0:
                        image_array[y + 1, x - 1] += quant_error * 3 / 16
                    
                    image_array[y + 1, x] += quant_error * 5 / 16
                    
                    if x + 1 < image_array.shape[1]:
                        image_array[y + 1, x + 1] += quant_error * 1 / 16
        
        # Convert back to PIL image
        dithered_image = Image.fromarray(image_array.astype('uint8'), 'L')
        
        # Convert to black and white
        return dithered_image.convert('1')



from airunner.gui.widgets.nodegraph.nodes.llm.llm_branch_node import (
    LLMBranchNode,
)
from airunner.gui.widgets.nodegraph.nodes.llm.gemma3_node import Gemma3Node

__all__ = ["LLMBranchNode", "Gemma3Node"]



import os
from queue import Queue
import time
import requests
from PySide6.QtCore import QObject, Signal
from airunner.enums import SignalCode
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.gui.windows.main.settings_mixin import SettingsMixin


DEFAULT_HF_ENDPOINT = "https://huggingface.co"


class DownloadWorker(
    MediatorMixin,
    SettingsMixin,
    QObject,
):
    progress = Signal(int, int)  # current, total
    finished = Signal()
    failed = Signal(Exception)
    queue = Queue()
    running = False
    is_cancelled = False

    def __init__(self, *args, **kwargs):
        super().__init__()

    def add_to_queue(self, data: tuple):
        self.queue.put(data)

    @staticmethod
    def get_size(url: str):
        try:
            response = requests.head(url, allow_redirects=True)
            size_kb = int(response.headers.get("content-length", 0))
            return size_kb
        except OverflowError:
            print(f"OverflowError when getting size for {url}")
            raise

    def cancel(self):
        self.is_cancelled = True

    def download(self):
        self.running = True

        while self.running:
            if self.queue.empty():
                time.sleep(0.1)
                continue

            path, file_name, file_path, callback = self.queue.get()
            if path == "" and file_name == "" and file_path == "":
                callback()
                return
            url = f"{DEFAULT_HF_ENDPOINT}/{path}/resolve/main/{file_name}?download=true".replace(
                " ", ""
            )
            self.emit_signal(SignalCode.CLEAR_DOWNLOAD_STATUS_BAR)
            self.emit_signal(
                SignalCode.SET_DOWNLOAD_STATUS_LABEL,
                {"message": f"Downloading {file_name}"},
            )

            file_name = os.path.join(file_path, file_name)
            file_name = os.path.expanduser(file_name)
            if not os.path.exists(os.path.dirname(file_name)):
                try:
                    os.makedirs(os.path.dirname(file_name), exist_ok=True)
                except FileExistsError:
                    pass

            if os.path.exists(file_name):
                self.emit_signal(
                    SignalCode.UPDATE_DOWNLOAD_LOG,
                    {"message": f"File already exists, skipping download"},
                )
                self.emit_signal(
                    SignalCode.DOWNLOAD_PROGRESS, {"current": 0, "total": 0}
                )
                self.finished.emit()
                continue

            size_kb = self.get_size(url)
            self.emit_signal(
                SignalCode.UPDATE_DOWNLOAD_LOG,
                {
                    "message": f"Downloading {url} of size {size_kb} KB to {file_name}"
                },
            )

            try:
                headers = {}
                with requests.get(
                    url, headers=headers, stream=True, allow_redirects=True
                ) as r:
                    r.raise_for_status()
                    dir_name = os.path.dirname(file_name)

                    if dir_name != "" and not os.path.exists(dir_name):
                        try:
                            os.makedirs(dir_name, exist_ok=True)
                        except FileExistsError:
                            pass

                    with open(file_name, "wb") as f:
                        for chunk in r.iter_content(chunk_size=8192):
                            if self.is_cancelled:
                                break
                            try:
                                f.write(chunk)
                            except OverflowError:
                                print(
                                    f"OverflowError when writing to {file_name}"
                                )
                                raise
                            self.emit_signal(
                                SignalCode.DOWNLOAD_PROGRESS,
                                {"current": f.tell(), "total": size_kb},
                            )
                        self.emit_signal(
                            SignalCode.UPDATE_DOWNLOAD_LOG,
                            {
                                "message": f"finished with download of {file_name}"
                            },
                        )
                        self.finished.emit()
            except Exception as e:
                print(f"Failed to download {url}")
                print(e)
                self.failed.emit(e)
                self.emit_signal(SignalCode.DOWNLOAD_COMPLETE)



from dataclasses import dataclass

@dataclass
class Rect:
    x: int
    y: int
    width: int
    height: int

    def left(self) -> int:
        """Return the x-coordinate of the left edge."""
        return self.x

    def top(self) -> int:
        """Return the y-coordinate of the top edge."""
        return self.y

    def translate(self, x: int, y: int):
        """Translate the rectangle by the given x and y offsets."""
        self.x += x
        self.y += y